"""Console script for covid19_outbreak_simulator."""
import argparse
import sys
import multiprocessing
import numpy as np
from io import StringIO
from tqdm import tqdm
from collections import defaultdict
from .simulator import Simulator
from .model import Params


def summarize_simulations(args):
    # write some summary information into standard output
    n_simulation = 0
    total_infection = 0
    total_infection_failed = 0
    total_infection_avoided = 0
    total_infection_ignored = 0
    total_show_symptom = 0
    total_removal = 0
    total_quarantine = 0
    total_reintegration = 0
    total_abort = 0
    #
    total_asym_infection = 0
    total_presym_infection = 0
    total_sym_infection = 0
    #
    n_no_outbreak = 0
    n_remaining_popsize = defaultdict(int)
    n_outbreak_duration = defaultdict(int)
    #
    n_num_infected_by_seed = defaultdict(int)
    n_first_infected_by_seed_on_day = defaultdict(int)
    n_seed_show_symptom_on_day = defaultdict(int)
    n_first_symptom_on_day = defaultdict(int)
    n_first_infection_on_day = defaultdict(int)
    #
    n_second_symptom_on_day = defaultdict(int)
    n_third_symptom_on_day = defaultdict(int)
    #
    with open(args.logfile) as lines:
        infection_from_seed_per_sim = defaultdict(int)
        infection_time_from_seed_per_sim = defaultdict(int)
        first_infection_day_per_sim = defaultdict(int)
        first_symptom_day_per_sim = defaultdict(int)
        second_symptom_per_sim = defaultdict(int)
        third_symptom_per_sim = defaultdict(int)

        if args.infectors:
            infectors = args.infectors
        else:
            infectors = ['0']

        total_popsize = sum(int(x.split('=')[-1]) for x in args.popsize)
        for line in lines:
            id, time, event, target, params = line.split('\t')
            if id == 'id':
                # skip header
                continue
            time = float(time)
            target = None if target == '.' else target
            params = params.split(',')
            #
            if event == 'INFECTION':
                total_infection += 1
                if any(x in params for x in [f'by={x}' for x in infectors]):
                    infection_from_seed_per_sim[id] += 1
                    if id not in infection_time_from_seed_per_sim:
                        infection_time_from_seed_per_sim[id] = time
                if time != 0 and id not in first_infection_day_per_sim:
                    first_infection_day_per_sim[id] = time
                for param in params:
                    if param.startswith('r_asym='):
                        total_asym_infection += int(param[7:])
                    elif param.startswith('r_presym='):
                        total_presym_infection += int(param[9:])
                    elif param.startswith('r_sym='):
                        total_sym_infection += int(param[6:])
            elif event == 'INFECTION_FAILED':
                total_infection_failed += 1
            elif event == 'INFECTION_AVOIDED':
                total_infection_avoided += 1
            elif event == 'INFECTION_IGNORED':
                total_infection_ignored += 1
            elif event == 'SHOW_SYMPTOM':
                total_show_symptom += 1
                if target in infectors:
                    n_seed_show_symptom_on_day[int(time) + 1] += 1
                if args.pre_quarantine and time < float(args.pre_quarantine[0]):
                    pass
                elif id not in first_symptom_day_per_sim:
                    first_symptom_day_per_sim[id] = time
                elif id not in second_symptom_per_sim:
                    second_symptom_per_sim[
                        id] = time - first_symptom_day_per_sim[id]
                elif id not in third_symptom_per_sim:
                    third_symptom_per_sim[
                        id] = time - second_symptom_per_sim[id]
            elif event == 'REMOVAL':
                total_removal += 1
            elif event == 'QUARANTINE':
                total_quarantine += 1
            elif event == 'REINTEGRATION':
                total_reintegration += 1
            elif event == 'ABORT':
                total_abort += 1
            elif event == 'END':
                n_simulation += 1
                n_remaining_popsize[target] += 1
                if not args.pre_quarantine:
                    n_outbreak_duration[0 if time == 0 else int(time) + 1] += 1
                else:
                    n_outbreak_duration[
                        0 if time <= float(args.pre_quarantine[0]) else
                        int(time - float(args.pre_quarantine[0])) + 1] += 1
                if target == total_popsize and id not in first_symptom_day_per_sim:
                    n_no_outbreak += 1
            else:
                raise ValueError(f'Unrecognized event {event}')
    # summarize
    for v in infection_from_seed_per_sim.values():
        n_num_infected_by_seed[v] += 1
    #
    for v in infection_time_from_seed_per_sim.values():
        if args.pre_quarantine:
            n_first_infected_by_seed_on_day[int(v -
                                                float(args.pre_quarantine[0])) +
                                            1] += 1
        else:
            n_first_infected_by_seed_on_day[int(v) + 1] += 1
    n_first_infected_by_seed_on_day[0] = n_simulation - sum(
        n_first_infected_by_seed_on_day.values())
    #
    for v in first_infection_day_per_sim.values():
        if args.pre_quarantine:
            n_first_infection_on_day[int(v - float(args.pre_quarantine[0])) +
                                     1] += 1
        else:
            n_first_infection_on_day[int(v) + 1] += 1
    n_first_infection_on_day[0] = n_simulation - sum(
        n_first_infection_on_day.values())
    #
    n_seed_show_symptom_on_day[0] = n_simulation - sum(
        n_seed_show_symptom_on_day.values())
    #
    for v in first_symptom_day_per_sim.values():
        if args.pre_quarantine:
            n_first_symptom_on_day[int(v - float(args.pre_quarantine[0])) +
                                   1] += 1
        else:
            n_first_symptom_on_day[int(v) + 1] += 1
    n_first_symptom = sum(n_first_symptom_on_day.values())
    #
    for v in second_symptom_per_sim.values():
        n_second_symptom_on_day[int(v) + 1] += 1
    n_second_symptom = sum(n_second_symptom_on_day.values())
    #
    for v in third_symptom_per_sim.values():
        n_third_symptom_on_day[int(v) + 1] += 1
    n_third_symptom = sum(n_third_symptom_on_day.values())
    #
    # print
    print(f'logfile\t{args.logfile}')
    print(f'popsize\t{" ".join(args.popsize)}')
    print(f'keep_symptomatic\t{"yes" if args.keep_symptomatic else "no"}')
    if args.symptomatic_r0:
        print(f'symptomatic_r0\t{" - ".join(args.symptomatic_r0)}')
    if args.asymptomatic_r0:
        print(f'asymptomatic_r0\t{" - ".join(args.asymptomatic_r0)}')
    if args.incubation_period:
        print(
            f'incubation_period\t{args.incubation_period[0]}({args.incubation_period[1]}, {args.incubation_period[2]}) '
            + ' '.join(args.incubation_period[3:]))
    if args.pre_quarantine is None:
        print(f'pre_quarantine\tno')
    else:
        print(f'pre_quarantine\t{" ".join(args.pre_quarantine)}')
    if args.interval == 1 / 24:
        print(f'interval\t1 hour')
    elif args.interval == 1:
        print(f'interval\t1 day')
    else:
        print(f'interval\t{args.interval:.2f} day')
    if args.prop_asym_carriers:
        if len(args.prop_asym_carriers) == 1:
            print(f'prop_asym_carriers\t{args.prop_asym_carriers[0]*100:.1f}%')
        else:
            print(
                f'prop_asym_carriers\t{args.prop_asym_carriers[0]*100:.1f}% to {args.prop_asym_carriers[1]*100:.1f}%'
            )
    else:
        print(f'prop_asym_carriers\t10% to 40%')
    print(f'allow_lead_time\t{"yes" if args.allow_lead_time else "no"}')
    print(f'n_simulation\t{n_simulation}')
    print(f'total_infection\t{total_infection}')
    print(f'total_infection_failed\t{total_infection_failed}')
    print(f'total_infection_avoided\t{total_infection_avoided}')
    print(f'total_infection_ignored\t{total_infection_ignored}')
    print(f'total_show_symptom\t{total_show_symptom}')
    print(f'total_removal\t{total_removal}')
    print(f'total_quarantine\t{total_quarantine}')
    print(f'total_reintegration\t{total_reintegration}')
    print(f'total_abort\t{total_abort}')
    print(f'total_asym_infection\t{total_asym_infection}')
    print(f'total_presym_infection\t{total_presym_infection}')
    print(f'total_sym_infection\t{total_sym_infection}')

    for num in sorted(n_remaining_popsize.keys()):
        print(f'n_remaining_popsize_{num}\t{n_remaining_popsize[num]}')
    # no outbreak is defined as final population size
    print(f'n_no_outbreak\t{n_no_outbreak}')
    for day in sorted(n_outbreak_duration.keys()):
        print(f'n_outbreak_duration_{day}\t{n_outbreak_duration[day]}')
    for num in sorted(n_num_infected_by_seed.keys()):
        print(f'n_num_infected_by_seed_{num}\t{n_num_infected_by_seed[num]}')
    for day in sorted(n_first_infected_by_seed_on_day.keys()):
        if day == 0:
            print(
                f'n_no_infected_by_seed\t{n_first_infected_by_seed_on_day[day]}'
            )
        else:
            print(
                f'n_first_infected_by_seed_on_day_{day}\t{n_first_infected_by_seed_on_day[day]}'
            )
    for day in sorted(n_seed_show_symptom_on_day.keys()):
        if day == 0:
            print(f'n_seed_show_no_symptom\t{n_seed_show_symptom_on_day[day]}')
        else:
            print(
                f'{day}_n_seed_show_symptom_on_day\t{n_seed_show_symptom_on_day[day]}'
            )
    for day in sorted(n_first_infection_on_day.keys()):
        if day == 0:
            print(f'n_no_first_infection\t{n_first_infection_on_day[day]}')
        else:
            print(
                f'n_first_infection_on_day_{day}\t{n_first_infection_on_day[day]}'
            )
    print(f'n_first_symptom\t{n_first_symptom}')
    for day in sorted(n_first_symptom_on_day.keys()):
        print(f'n_first_symptom_on_day_{day}\t{n_first_symptom_on_day[day]}')
    print(f'n_second_symptom\t{n_second_symptom}')
    for day in sorted(n_second_symptom_on_day.keys()):
        print(f'n_second_symptom_on_day_{day}\t{n_second_symptom_on_day[day]}')
    print(f'n_third_symptom\t{n_third_symptom}')
    for day in sorted(n_third_symptom_on_day.keys()):
        print(f'n_third_symptom_on_day_{day}\t{n_third_symptom_on_day[day]}')


class Worker(multiprocessing.Process):

    def __init__(self, task_queue, result_queue, args):
        multiprocessing.Process.__init__(self)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self.params = Params(args)
        self.simu_args = args

    def run(self):
        # set random seed to a random number
        np.random.seed()
        while True:
            id = self.task_queue.get()
            if id is None:
                self.task_queue.task_done()
                break
            with StringIO() as logger:
                logger.id = id
                simu = Simulator(
                    params=self.params, logger=logger, simu_args=self.simu_args)
                simu.simulate(id)
                self.task_queue.task_done()
                self.result_queue.put(logger.getvalue())


def parse_args(args=None):
    parser = argparse.ArgumentParser('outbreak_simulator')
    parser.add_argument(
        '--popsize',
        default=['64'],
        nargs='+',
        help='''Size of the population, including the infector that
        will be introduced at the beginning of the simulation. It should be specified
        as a single number, or a serial of name=size values for different groups. For
        example "--popsize nurse=10 patient=100". The names will be used for setting
        group specific parameters. The IDs of these individuals will be nurse0, nurse1
        etc. ''')
    parser.add_argument(
        '--susceptibility',
        nargs='+',
        help='''Weight of susceptibility. The default value is 1, meaning everyone is
            equally susceptible. With options such as
            "--susceptibility nurse=1.2 patients=0.8" you can give weights to different
            groups of people so that they have higher or lower probabilities to be
            infected.''')
    parser.add_argument(
        '--symptomatic-r0',
        nargs='+',
        help='''Production number of symptomatic infectors, should be specified as a single
            fixed number, or a range, and/or multipliers for different groups such as
            A=1.2. For example "--symptomatic-r0 1.4 2.8 nurse=1.2" means a general R0
            ranging from 1.4 to 2.8, while nursed has a range from 1.4*1.2 and 2.8*1.2.'''
    )
    parser.add_argument(
        '--asymptomatic-r0',
        nargs='+',
        help='''Production number of asymptomatic infectors, should be specified as a single
            fixed number, or a range and/or multipliers for different groups''')
    parser.add_argument(
        '--incubation-period',
        nargs='+',
        help='''Incubation period period, should be specified as "lognormal" followed by two
            numbers as mean and sigma, or "normal" followed by mean and sd, and/or
            multipliers for different groups. Default to "lognormal 1.621 0.418"'''
    )
    parser.add_argument(
        '--repeats',
        default=10000,
        type=int,
        help='''Number of replicates to simulate. An ID starting from
              1 will be assinged to each replicate and as the first columns
              in the log file.''')
    parser.add_argument(
        '--keep-symptomatic',
        action='store_true',
        help='Keep affected individuals in the population')
    parser.add_argument(
        '--pre-quarantine',
        nargs='*',
        help='''Days of self-quarantine before introducing infector to the group.
            The simulation will be aborted if the infector shows symptom before
            introduction. If you quarantine multiple people or specified named
            groups, you will need to append the IDs to the parameter (e.g.
            --pre-quarantine day nurse1 nurse2''')
    parser.add_argument(
        '--infectors',
        nargs='*',
        help='''Infectees to introduce to the population, default to '0'. If you
            would like to introduce multiple infectees to the population, or if
            you have named groups, you will have to specify the IDs of carrier
            such as --infectors nurse1 nurse2''')
    parser.add_argument(
        '--interval',
        default=1 / 24,
        help='Interval of simulation, default to 1/24, by hour')
    parser.add_argument('--logfile', default='simulation.log', help='logfile')

    parser.add_argument(
        '--prop-asym-carriers',
        nargs='*',
        help='''Proportion of asymptomatic cases. You can specify a fix number, or two
        numbers as the lower and higher CI (95%%) of the proportion. Default to 0.10 to 0.40.
        Multipliers can be specified to set proportion of asymptomatic carriers
        for particular groups.''')
    parser.add_argument(
        '--allow-lead-time',
        action='store_true',
        help='''The seed carrier will be asumptomatic but always be at the beginning
            of incurbation time. If allow lead time is set to True, the carrier will
            be anywhere in his or her incubation period.''')
    parser.add_argument(
        '--analyze-existing-logfile',
        action='store_true',
        help='''Analyze an existing logfile, useful for updating the summarization
            procedure or uncaptured output. ''')
    parser.add_argument(
        '-j',
        '--jobs',
        type=int,
        help='Number of process to use for simulation. Default to number of CPU cores.'
    )
    return parser.parse_args(args)


def main(args=None):
    """Console script for covid19_outbreak_simulator."""
    args = parse_args(args)
    if not args.analyze_existing_logfile:
        tasks = multiprocessing.JoinableQueue()
        results = multiprocessing.Queue()

        if not args.jobs:
            args.jobs = multiprocessing.cpu_count()

        workers = [Worker(tasks, results, args) for i in range(args.jobs)]
        for worker in workers:
            worker.start()

        with open(args.logfile, 'w') as logger:
            logger.write('id\ttime\tevent\ttarget\tparams\n')
            for i in range(args.repeats):
                tasks.put(i + 1)
            for i in range(args.jobs):
                tasks.put(None)
            #
            for i in tqdm(range(args.repeats)):
                result = results.get()
                logger.write(result)

    summarize_simulations(args)
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import pytest

from covid19_outbreak_simulator.model import Params, Model
from covid19_outbreak_simulator.simulator import Simulator, Individual
from argparse import Namespace


@pytest.fixture
def params():
    return Params()


@pytest.fixture
def default_model():
    return Model(Params())


@pytest.fixture
def logger():
    with open('test.log', 'w') as logger:
        logger.id = 1
        yield logger


@pytest.fixture
def simulator(params, logger):
    simu_args = Namespace()
    simu_args.popsize = 64
    simu_args.repeat = 1
    simu_args.keep_symptomatic = True
    simu_args.pre_quarantine = None

    return Simulator(params=params, logger=logger, simu_args=simu_args)


@pytest.fixture
def individual(default_model, logger):
    return Individual(0, 'group', 0.8, default_model, logger)


@pytest.fixture
def individual_factory(default_model, logger):
    return lambda id: Individual(id, 'group 2', 1.2, default_model, logger)


from enum import Enum

import random
import numpy as np
from numpy.random import choice
from collections import defaultdict

from .model import Model


class Individual(object):

    def __init__(self, id, group, susceptibility, model, logger):
        self.id = id
        self.model = model
        self.group = group
        self.susceptibility = susceptibility
        self.logger = logger

        self.infected = None
        self.quarantined = None
        self.n_infectee = 0

        self.r0 = None
        self.incubation_period = None

    def quarantine(self, till):
        self.quarantined = till
        return [Event(till, EventType.REINTEGRATION, self, logger=self.logger)]

    def reintegrate(self):
        self.quarantined = None
        return []

    def symptomatic_infect(self, time, **kwargs):
        self.infected = time
        self.r0 = self.model.draw_random_r0(symptomatic=True, group=self.group)
        self.incubation_period = self.model.draw_random_incubation_period(
            group=self.group)

        by = kwargs.get('by', None)

        if by is None and 'allow_lead_time' in kwargs:
            # this is the first infection, the guy should be asymptomatic, but
            # could be anywhere in his incubation period
            lead_time = np.random.uniform(0, self.incubation_period)
        else:
            lead_time = 0

        keep_symptomatic = kwargs.get('keep_symptomatic', False)

        # REMOVAL ...
        evts = [
            Event(
                time + self.incubation_period - lead_time,
                EventType.SHOW_SYMPTOM,
                self,
                logger=self.logger)
        ]
        if not keep_symptomatic:
            if self.quarantined and time + self.incubation_period - lead_time < self.quarantined:
                # scheduling ABORT
                evts.append(
                    Event(
                        time + self.incubation_period - lead_time,
                        EventType.ABORT,
                        self,
                        logger=self.logger))
            else:
                evts.append(
                    # scheduling REMOVAL
                    Event(
                        time + self.incubation_period - lead_time,
                        EventType.REMOVAL,
                        self,
                        logger=self.logger))
        #
        x_grid, trans_prob = self.model.get_symptomatic_transmission_probability(
            self.incubation_period, self.r0)
        # infect only before removal
        if keep_symptomatic:
            x_before = x_grid
        else:
            x_before = [
                x for x in x_grid if x < self.incubation_period - lead_time
            ]
        infected = np.random.binomial(1, trans_prob[:len(x_before)],
                                      len(x_before))
        presymptomatic_infected = [
            xx for xx, ii in zip(x_before, infected)
            if ii and xx < self.incubation_period
        ]
        symptomatic_infected = [
            xx for xx, ii in zip(x_before, infected)
            if ii and xx >= self.incubation_period
        ]
        if self.quarantined:
            for idx, x in enumerate(x_before):
                if time + x < self.quarantined and infected[idx] != 0:
                    evts.append(
                        Event(
                            time + x,
                            EventType.INFECTION_AVOIDED,
                            self.id,
                            logger=self.logger,
                            by=self))
                    infected[idx] = 0
        #
        for x, infe in zip(x_before, infected):
            if infe:
                evts.append(
                    Event(
                        time + x,
                        EventType.INFECTION,
                        None,
                        logger=self.logger,
                        by=self))

        if by:
            by.n_infectee += 1
            params = [f'by={by.id}']
        elif lead_time:
            params = [f'leadtime={lead_time:.2f}']
        else:
            params = []
        #
        params.extend([
            f'r0={self.r0:.2f}',
            f'r={sum(infected)}',
            f'r_presym={len(presymptomatic_infected)}',
            f'r_sym={len(symptomatic_infected)}',
            f'incu={self.incubation_period:.2f}',
        ])
        self.logger.write(
            f'{self.logger.id}\t{time:.2f}\t{EventType.INFECTION.name}\t{self.id}\t{",".join(params)}\n'
        )
        return evts

    def asymptomatic_infect(self, time, **kwargs):
        self.infected = time
        self.r0 = self.model.draw_random_r0(symptomatic=False)
        self.incubation_period = -1

        by = kwargs.get('by',)

        if by is None and 'allow_lead_time' in kwargs:
            # this is the first infection, the guy should be asymptomatic, but
            # could be anywhere in his incubation period
            lead_time = np.random.uniform(0, 10)
        else:
            lead_time = 0

        # REMOVAL ...
        evts = []
        #
        x_grid, trans_prob = self.model.get_asymptomatic_transmission_probability(
            self.r0)

        if lead_time > 0:
            idx = int(lead_time / self.model.params.simulation_interval)
            trans_prob = trans_prob[idx:]
            x_grid = x_grid[idx:]
            x_grid = x_grid - x_grid[0]

        # infect only before removal
        infected = np.random.binomial(1, trans_prob, len(x_grid))
        asymptomatic_infected = sum(infected)
        if self.quarantined:
            for idx, x in enumerate(x_grid):
                if time + x < self.quarantined and infected[idx] != 0:
                    evts.append(
                        Event(
                            time + x,
                            EventType.INFECTION_AVOIDED,
                            self.id,
                            logger=self.logger,
                            by=self))
                    infected[idx] = 0
        #
        for x, infe in zip(x_grid, infected):
            if infe:
                evts.append(
                    Event(
                        time + x,
                        EventType.INFECTION,
                        None,
                        logger=self.logger,
                        by=self))

        if by:
            by.n_infectee += 1
            params = [f'by={by.id}']
        elif lead_time > 0:
            params = [f'leadtime={lead_time:.2f}']
        else:
            params = []
        #
        params.extend([
            f'r0={self.r0:.2f}', f'r={asymptomatic_infected}',
            f'r_asym={asymptomatic_infected}'
        ])
        self.logger.write(
            f'{self.logger.id}\t{time:.2f}\t{EventType.INFECTION.name}\t{self.id}\t{",".join(params)}\n'
        )
        return evts

    def infect(self, time, **kwargs):
        if self.infected is not None:
            self.logger.write(
                f'{self.logger.id}\t{time:.2f}\t{EventType.INFECTION_IGNORED.name}\t{self.id}\tby={kwargs["by"].id}\n'
            )
            return []

        if self.model.draw_is_asymptomatic():
            return self.asymptomatic_infect(time, **kwargs)
        else:
            return self.symptomatic_infect(time, **kwargs)


class EventType(Enum):
    # Infection
    INFECTION = 1
    # infection failed due to perhaps no more people to infect
    INFECTION_FAILED = 2
    # infection event happens during quarantine
    INFECTION_AVOIDED = 3
    # infection event happens to an infected individual
    INFECTION_IGNORED = 4

    # removal of individual showing symptom
    SHOW_SYMPTOM = 5
    REMOVAL = 6
    # quarantine individual given a specified time
    QUARANTINE = 7
    # reintegrate individual to the population (release from quarantine)
    REINTEGRATION = 8

    # abort simulation, right now due to infector showing symptoms during quarantine
    ABORT = 9
    # end of simulation
    END = 10


class Event(object):
    '''
    Events that happen during the simulation.
    '''

    def __init__(self, time, action, target=None, logger=None, **kwargs):
        self.time = time
        self.action = action
        self.target = target
        self.logger = logger
        self.kwargs = kwargs

    def apply(self, population, simu_args):
        if self.action == EventType.INFECTION:
            if self.target is not None:
                selected = self.target
            elif simu_args.susceptibility:
                # select one non-quarantined indivudal to infect
                ids = [(id, ind.susceptibility)
                       for id, ind in population.items()
                       if (not self.target or id != self.target.id) and
                       not ind.quarantined]

                if not ids:
                    self.logger.write(
                        f'{self.logger.id}\t{self.time:.2f}\t{EventType.INFECTION_FAILED.name}\t{self.target.id}\tby={self.kwargs["by"].id}\n'
                    )
                    return []
                weights = np.array([x[1] for x in ids])
                weights = weights / sum(weights)
                selected = ids[choice(len(ids), 1, p=weights)[0]][0]
            else:
                # select one non-quarantined indivudal to infect
                ids = [
                    id for id, ind in population.items()
                    if (not self.target or id != self.target.id) and
                    not ind.quarantined
                ]
                if not ids:
                    self.logger.write(
                        f'{self.logger.id}\t{self.time:.2f}\t{EventType.INFECTION_FAILED.name}\t{self.target.id}\tby={self.kwargs["by"].id}\n'
                    )
                    return []
                selected = random.choice(ids)
            return population[selected].infect(
                self.time,
                keep_symptomatic=simu_args.keep_symptomatic,
                allow_lead_time=simu_args.allow_lead_time,
                **self.kwargs)
        elif self.action == EventType.QUARANTINE:
            self.logger.write(
                f'{self.logger.id}\t{self.time:.2f}\t{EventType.QUARANTINE.name}\t{self.target.id}\ttill={self.kwargs["till"]:.2f}\n'
            )
            return population[self.target.id].quarantine(**self.kwargs)
        elif self.action == EventType.REINTEGRATION:
            self.logger.write(
                f'{self.logger.id}\t{self.time:.2f}\t{EventType.REINTEGRATION.name}\t{self.target.id}\t.\n'
            )
            return population[self.target.id].reintegrate(**self.kwargs)
        elif self.action == EventType.INFECTION_AVOIDED:
            self.logger.write(
                f'{self.logger.id}\t{self.time:.2f}\t{EventType.INFECTION_AVOIDED.name}\t.\tby={self.kwargs["by"].id}\n'
            )
            return []
        elif self.action == EventType.SHOW_SYMPTOM:
            self.logger.write(
                f'{self.logger.id}\t{self.time:.2f}\t{EventType.SHOW_SYMPTOM.name}\t{self.target.id}\t.\n'
            )
            return []
        elif self.action == EventType.REMOVAL:
            population.pop(self.target.id)
            self.logger.write(
                f'{self.logger.id}\t{self.time:.2f}\t{EventType.REMOVAL.name}\t{self.target.id}\tpopsize={len(population)}\n'
            )
            return []
        else:
            raise RuntimeError(f'Unrecognized action {self.action}')

    def __str__(self):
        return f'{self.action.name}_{self.target.id if self.target else ""}_at_{self.time:.2f}'


class Simulator(object):

    def __init__(self, params, logger, simu_args):
        self.logger = logger
        self.simu_args = simu_args
        self.params = params
        self.model = None

    def simulate(self, id):
        #
        # get proportion of asymptomatic
        #
        self.model = Model(self.params)
        self.model.draw_prop_asym_carriers()

        # collection of individuals
        population = {}
        idx = 0
        for ps in self.simu_args.popsize:
            if '=' in ps:
                # this is named population size
                name, sz = ps.split('=', 1)

            else:
                name = ''
                sz = ps
            try:
                sz = int(sz)
            except Exception:
                raise ValueError(
                    f'Named population size should be name=int: {ps} provided')

            population.update({
                name + str(idx): Individual(
                    name + str(idx),
                    group=name,
                    susceptibility=getattr(self.model.params,
                                           f'susceptibility_multiplier_{name}',
                                           1),
                    model=self.model,
                    logger=self.logger) for idx in range(idx, idx + sz)
            })

        events = defaultdict(list)
        self.logger.id = id

        # quanrantine the first person if args.pre-quarantine > 0
        if self.simu_args.pre_quarantine:
            try:
                till = float(self.simu_args.pre_quarantine[0])
            except Exception:
                raise ValueError(
                    f'The first value of pre_quarantine should be days (a float number): {self.simu_args.pre_quarantine[0]} provided'
                )
            IDs = ['0'] if len(self.simu_args.pre_quarantine
                              ) == 1 else self.simu_args.pre_quarantine[1:]
            for ID in IDs:
                if ID not in population:
                    raise ValueError(f'Invalid ID to quanrantine {ID}')
                events[0].append(
                    Event(
                        0,
                        EventType.QUARANTINE,
                        population[ID],
                        logger=self.logger,
                        till=till))

        infectors = [
            '0'
        ] if not self.simu_args.infectors else self.simu_args.infectors
        for infector in infectors:
            if infector not in population:
                raise ValueError(f'Invalid ID for carrier {infector}')
            # infect the first person
            events[0].append(
                Event(
                    0, EventType.INFECTION, target=infector,
                    logger=self.logger))

        while True:
            # find the latest event
            time = min(events.keys())

            new_events = []
            aborted = False
            # processing events
            for evt in events[time]:
                if evt.action == EventType.ABORT:
                    self.logger.write(
                        f'{self.logger.id}\t{time:.2f}\t{EventType.ABORT.name}\t{evt.target.id}\tpopsize={len(population)}\n'
                    )
                    aborted = True
                    break
                # event triggers new event
                new_events.extend(evt.apply(population, self.simu_args))
            events.pop(time)
            #
            for evt in new_events:
                # print(f'ADDING\t{evt}')
                events[evt.time].append(evt)

            if not events or aborted:
                break
            # if self.simu_args.keep_symptomatic and all(
            #         x.infected for x in population.values()):
            #     break
        self.logger.write(
            f'{self.logger.id}\t{time:.2f}\t{EventType.END.name}\t{len(population)}\tpopsize={len(population)},prop_asym={self.model.params.prop_asym_carriers:.3f}\n'
        )



import random
import re

import numpy as np
from scipy.stats import norm
from scipy.optimize import bisect


class Params:

    def __init__(self, args=None):
        # args should be args returned from argparser
        self.params = set([
            'simulation_interval', 'prop_asym_carriers', 'incubation_period',
            'symptomatic_r0', 'asymptomatic_r0', 'susceptibility'
        ])
        self.groups = {}
        self.set_params(args)

    def set(self, param, prop, value):
        if param not in self.params:
            raise ValueError(f'Unrecgonzied parameter {param}')
        if prop is None or prop == 'self':
            setattr(self, param, value)
        elif prop in ('loc', 'low', 'high', 'mean', 'sigma', 'scale'):
            setattr(self, f'{param}_{prop}', value)
        elif prop.startswith('multiplier_'):
            group = prop[11:]
            if group not in self.groups:
                raise ValueError(
                    f'Group {group} does not exist. Available groups are: {", ".join(self.groups.keys())}'
                )
            if value < 0:
                raise ValueError(
                    f'Multiplier should be positive {value} specified')
            setattr(self, f'{param}_{prop}', value)
        elif re.match('quantile_(.*)', prop):
            lq = float(re.match('quantile_(.*)', prop)[1]) / 100
            loc = getattr(self, f'{param}_loc')
            sd = bisect(
                lambda x: norm.cdf(value, loc=loc, scale=x) - lq,
                a=0.00001,
                b=25)
            setattr(self, f'{param}_scale', sd)
        else:
            raise ValueError(f'Unrecognized property {prop}')

    def check_id(self, ID):
        if ID.isnumeric():
            if '' not in self.groups:
                raise ValueError(f'Invalid ID {ID}: No default group.')
            if int(ID) >= self.groups[''] or int(ID) < 0:
                raise ValueError(f'Invalid ID {ID}: Index out of range.')
        else:
            found = False
            for name, size in self.groups.items():
                if ID.startswith(name) and ID[len(name):].isnumeric():
                    idx = int(ID[len(name):])
                    if idx >= self.groups[name] or idx < 0:
                        raise ValueError(
                            f'Invalid ID {ID}: Index out of range.')
                    found = True
                    break
            if not found:
                raise ValueError(f'Invalid ID {ID}')

    def set_params(self, args):
        # set some default values first
        self.set('simulation_interval', 'self',
                 args.interval if args else 1 / 24)
        self.set('prop_asym_carriers', 'loc', 0.25)
        self.set('prop_asym_carriers', 'quantile_2.5', 0.1)
        self.set('symptomatic_r0', 'low', 1.4)
        self.set('symptomatic_r0', 'high', 2.8)
        self.set('asymptomatic_r0', 'low', 0.28)
        self.set('asymptomatic_r0', 'high', 0.56)
        self.set('incubation_period', 'mean', 1.621)
        self.set('incubation_period', 'sigma', 0.418)

        if not args:
            return
        if args.popsize:
            for ps in args.popsize:
                if '=' in ps:
                    name, size = ps.split('=', 1)
                    if name[-1].isnumeric():
                        raise ValueError(
                            f'Group names that end with numbers are not allowed. {name}'
                        )
                else:
                    name = ''
                    size = int(ps)
                try:
                    size = int(size)
                except Exception:
                    raise ValueError(f'Incorrect popsize {ps}')
                if name in self.groups:
                    raise ValueError(
                        f'Group "{name}" has been specified before')
                self.groups[name] = size

        # do a favor to check parameters, but does not really save parameters here
        if args.pre_quarantine:
            try:
                float(args.pre_quarantine[0])
            except Exception:
                raise ValueError(
                    'The first parameter of pre_quarantine should be a number')
            for ID in args.pre_quarantine[1:]:
                self.check_id(ID)

        # do a favor to check parameters, but does not really save parameters here
        if args.infectors:
            for ID in args.infectors:
                self.check_id(ID)

        # modify from command line args
        if args.symptomatic_r0:
            # if asymptomatic_r0 is specified, it should a number of a range...
            pars = [x for x in args.symptomatic_r0 if '=' not in x]
            if len(pars) == 1:
                try:
                    self.set('symptomatic_r0', 'low', float(pars[0]))
                    self.set('symptomatic_r0', 'high', float(pars[0]))
                except Exception:
                    raise ValueError(
                        f'The symptomatic_r0 should be a float number, if it is not a multiplier for groups: {pars[0]} provided'
                    )
            elif len(pars) == 2:
                try:
                    self.set('symptomatic_r0', 'low', float(pars[0]))
                except Exception as e:
                    raise ValueError(
                        f'The symptomatic_r0 should be a float number, if it is not a multiplier for groups: {pars[0]} provided'
                    )
                try:
                    self.set('symptomatic_r0', 'high', float(pars[1]))
                except Exception as e:
                    raise ValueError(
                        f'The symptomatic_r0 should be a float number, if it is not a multiplier for groups: {pars[1]} provided'
                    )
            elif len(pars) > 2:
                raise ValueError(
                    f'The symptomatic_r0 should be one or two float number.')
            #
            for multiplier in [x for x in args.symptomatic_r0 if '=' in x]:
                name, value = multiplier.split('=', 1)
                try:
                    value = float(value)
                except Exception:
                    raise ValueError(
                        f'Multiplier should have format name=float_value: {multiplier} provided'
                    )
                self.set('symptomatic_r0', f'multiplier_{name}', value)
        if args.asymptomatic_r0:
            # if asymptomatic_r0 is specified, it should a number of a range...
            pars = [x for x in args.asymptomatic_r0 if '=' not in x]
            if len(pars) == 1:
                try:
                    self.set('asymptomatic_r0', 'low', float(pars[0]))
                    self.set('asymptomatic_r0', 'high', float(pars[0]))
                except Exception:
                    raise ValueError(
                        f'The asymptomatic_r0 should be a float number, if it is not a multiplier for groups: {pars[0]} provided'
                    )
            elif len(pars) == 2:
                try:
                    self.set('asymptomatic_r0', 'low', float(pars[0]))
                except Exception as e:
                    raise ValueError(
                        f'The asymptomatic_r0 should be a float number, if it is not a multiplier for groups: {pars[0]} provided'
                    )
                try:
                    self.set('asymptomatic_r0', 'high', float(pars[1]))
                except Exception as e:
                    raise ValueError(
                        f'The asymptomatic_r0 should be a float number, if it is not a multiplier for groups: {pars[1]} provided'
                    )
            elif len(pars) > 2:
                raise ValueError(
                    f'The asymptomatic_r0 should be one or two float number.')
            #
            for multiplier in [x for x in args.asymptomatic_r0 if '=' in x]:
                name, value = multiplier.split('=', 1)
                try:
                    value = float(value)
                except Exception:
                    raise ValueError(
                        f'Multiplier should have format name=float_value: {multiplier} provided'
                    )
                self.set('asymptomatic_r0', f'multiplier_{name}', value)
        if args.incubation_period:
            pars = [
                x for x in args.incubation_period
                if x in ('normal', 'lognormal') or '=' not in x
            ]
            if pars:
                if len(pars) < 3:
                    raise ValueError(
                        f'Parameter incubation period requires aat least three values: {len(args.incubation_period)} provided'
                    )
                if pars[0] not in ('normal', 'lognormal'):
                    raise ValueError(
                        f'Only normal or lognormal distribution for incubation period is supported. {pars[0]} provided'
                    )
                try:
                    self.set('incubation_period',
                             'mean' if pars[0] == 'lognormal' else 'loc',
                             float(pars[1]))
                except Exception:
                    raise ValueError(
                        f'Second parameter of incubation_period should be a float number: {pars[1]} provided'
                    )
                try:
                    self.set('incubation_period',
                             'sigma' if pars[0] == 'lognormal' else 'scale',
                             float(pars[2]))
                except Exception:
                    raise ValueError(
                        f'Third parameter of lognormal incubation_period should be a float number: {pars[2]} provided'
                    )
            # multipliers
            for multiplier in [x for x in args.incubation_period if '=' in x]:
                name, value = multiplier.split('=', 1)
                try:
                    value = float(value)
                except Exception:
                    raise ValueError(
                        f'Multiplier should have format name=float_value: {multiplier} provided'
                    )
                self.set('incubation_period', f'multiplier_{name}', value)
        if args.susceptibility:
            for multiplier in args.susceptibility:
                if '=' not in multiplier:
                    raise ValueError(
                        f'Susceptibility has to be specified as name=weight: {multiplier} specified.'
                    )
                name, value = multiplier.split('=', 1)
                try:
                    value = float(value)
                except Exception:
                    raise ValueError(
                        f'Multiplier should have format name=float_value: {multiplier} provided'
                    )
                self.set('susceptibility', f'multiplier_{name}', value)
        if args.prop_asym_carriers:
            try:
                pars = [
                    float(x) for x in args.prop_asym_carriers if '=' not in x
                ]
            except Exception:
                raise ValueError(
                    f'Paramter prop-asym-carriers expect one or two float value or multipliers: {" ".join(args.prop_asym_carriers)} provided'
                )
            if len(pars) == 1:
                self.set('prop_asym_carriers', 'loc', float(pars[0]))
                self.set('prop_asym_carriers', 'scale', 0)
            elif len(pars) == 2:
                if pars[0] > pars[1]:
                    raise ValueError(
                        f'Proportions for parameter prop-asym-carriers should be incremental.'
                    )
                self.set('prop_asym_carriers', 'loc', (pars[0] + pars[1]) / 2)
                self.set('prop_asym_carriers', 'quantile_2.5', pars[0])
            elif len(pars) > 2:
                raise ValueError(
                    f'Parameter prop-asym-carriers accepts one or two numbers.')
            #
            for multiplier in [x for x in args.prop_asym_carriers if '=' in x]:
                name, value = multiplier.split('=', 1)
                try:
                    value = float(value)
                except Exception:
                    raise ValueError(
                        f'Multiplier should have format name=float_value: {multiplier} provided'
                    )
                self.set('prop_asym_carriers', f'multiplier_{name}', value)


class Model(object):

    sd_5 = bisect(lambda x: norm.cdf(10, loc=5, scale=x) - 0.995, a=0.001, b=5)
    sd_6 = bisect(lambda x: norm.cdf(14, loc=6, scale=x) - 0.975, a=0.001, b=5)

    def __init__(self, params):
        self.params = params
        self.params.prop_asym_carriers = None

    def draw_prop_asym_carriers(self, group=''):
        self.params.prop_asym_carriers = np.random.normal(
            loc=self.params.prop_asym_carriers_loc,
            scale=self.params.prop_asym_carriers_scale)
        return min(
            max(
                self.params.prop_asym_carriers * getattr(
                    self.params, f'prop_asym_carriers_multiplier_{group}', 1.0),
                0), 1)

    def draw_is_asymptomatic(self):
        return np.random.uniform(0, 1) < self.params.prop_asym_carriers

    def draw_random_r0(self, symptomatic, group=''):
        '''
        Reproduction number, drawn randomly between 1.4 and 2.8.
        '''
        if symptomatic:
            r0 = np.random.uniform(self.params.symptomatic_r0_low,
                                   self.params.symptomatic_r0_high)
            return r0 * getattr(self.params,
                                f'symptomatic_r0_multiplier_{group}', 1.0)
        else:
            r0 = np.random.uniform(self.params.asymptomatic_r0_low,
                                   self.params.asymptomatic_r0_high)
            return r0 * getattr(self.params,
                                f'asymptomatic_r0_multiplier_{group}', 1.0)

    def draw_random_incubation_period(self, group=''):
        '''
        Incubation period, drawn from a lognormal distribution.
        '''
        if hasattr(self.params, 'incubation_period_loc'):
            # if a normal distribution is specified
            ip = max(
                0,
                np.random.normal(
                    loc=self.params.incubation_period_loc,
                    scale=self.params.incubation_period_scale))
        else:
            ip = np.random.lognormal(
                mean=self.params.incubation_period_mean,
                sigma=self.params.incubation_period_sigma)
        return ip * getattr(self.params,
                            f'incubation_period_multiplier_{group}', 1.0)

    def get_symptomatic_transmission_probability(self, incu, R0):
        '''Transmission probability.
        incu
            incubation period in days (can be float)

        R0
            reproductive number, which is the expected number of infectees

        interval
            interval of simulation, default to 1/24, which is by hours

        returns

        x
            time point
        y
            probability of transmission for each time point
        '''
        # right side with 6 day interval
        incu = incu * 2 / 3
        dist_right = norm(incu, self.sd_6)

        # if there is no left-hand-side
        if incu <= self.params.simulation_interval:
            x = np.linspace(0, incu + 8,
                            int((incu + 8) / self.params.simulation_interval))
            y = dist_right.pdf(x)
        else:
            # left hand side with a incu day interval
            try:
                sd_left = bisect(
                    lambda x: norm.cdf(2 * incu, loc=incu, scale=x) - 0.99,
                    a=0.001,
                    b=15,
                    xtol=0.001)
            except:
                # if incubation period is zer0
                sd_left = 0.0
            dist_left = norm(incu, sd_left)
            scale = dist_right.pdf(incu) / dist_left.pdf(incu)

            x = np.linspace(0, incu + 8,
                            int((incu + 8) / self.params.simulation_interval))
            idx = int(incu / self.params.simulation_interval)
            y = np.concatenate(
                [dist_left.pdf(x[:idx]) * scale,
                 dist_right.pdf(x[idx:])])
        sum_y = sum(y)
        return x, y / sum(y) * R0

    def get_asymptomatic_transmission_probability(self, R0):
        '''Asymptomatic Transmission probability.
        R0
            reproductive number, which is the expected number of infectees

        interval
            interval of simulation, default to 1/24, which is by hours

        returns

        x
            time point
        y
            probability of transmission for each time point
        '''
        dist = norm(4.8, self.sd_5)
        x = np.linspace(0, 12, int(12 / self.params.simulation_interval))
        y = dist.pdf(x)
        sum_y = sum(y)
        return x, y / sum(y) * R0



import pytest
from itertools import product
from covid19_outbreak_simulator.simulator import Individual, EventType


def test_individual(default_model, logger):
    ind = Individual(0, 'group 1', 1.2, default_model, logger)

    assert ind.infected is None
    assert ind.quarantined is None
    assert ind.n_infectee == 0

    assert ind.r0 is None
    assert ind.incubation_period is None


def test_quarantine(individual):
    res = individual.quarantine(till=2)

    assert individual.quarantined == 2
    assert len(res) == 1
    assert res[0].action == EventType.REINTEGRATION
    assert res[0].time == 2
    assert res[0].target == individual

    res = individual.reintegrate()

    assert not res
    assert individual.quarantined is None


def test_infect(individual):
    individual.model.draw_prop_asym_carriers()
    res = individual.infect(5.0, by=None)


def test_infect_infected(individual_factory):
    ind = individual_factory(id=1)
    ind1 = individual_factory(id=3)
    ind.model.draw_prop_asym_carriers()
    ind.infected = 5.5

    res = ind.infect(5.0, by=ind1)

    assert not res


@pytest.mark.parametrize('by, keep_symptomatic, quarantined',
                         product([None, 1], [True, False], [True, False]))
def test_symptomatic_infect(individual_factory, by, keep_symptomatic,
                            quarantined):
    ind1 = individual_factory(id=1)
    ind2 = individual_factory(id=2)

    ind1.model.draw_prop_asym_carriers()
    if quarantined:
        ind2.quarantine(20)
    #
    res = ind1.symptomatic_infect(
        5.0, by=None if by is None else ind2, keep_symptomatic=keep_symptomatic)

    assert ind1.infected == 5.0
    assert ind1.r0 is not None
    assert ind1.incubation_period is not None


@pytest.mark.parametrize('by, keep_symptomatic, quarantined',
                         product([None, 1], [True, False], [True, False]))
def test_asymptomatic_infect(individual_factory, by, keep_symptomatic,
                             quarantined):
    ind1 = individual_factory(id=1)
    ind2 = individual_factory(id=2)

    ind1.model.draw_prop_asym_carriers()
    if quarantined:
        ind2.quarantine(20)
    #
    res = ind1.asymptomatic_infect(
        5.0, by=None if by is None else ind2, keep_symptomatic=keep_symptomatic)

    assert ind1.infected == 5.0
    assert ind1.r0 is not None
    assert ind1.incubation_period is not None



import os
import pytest
from covid19_outbreak_simulator.cli import parse_args, main
from covid19_outbreak_simulator.model import Params

def test_option_popsize():
    args = parse_args(['--popsize', '500'])
    params = Params(args)
    assert list(params.groups.keys()) == ['']

    args = parse_args(['--popsize', 'A=500', 'B=300'])
    params = Params(args)
    assert sorted(list(params.groups.keys())) == ['A', 'B']

    args = parse_args(['--popsize', 'A=500', 'B=300'])
    params = Params(args)
    assert sorted(list(params.groups.keys())) == ['A', 'B']

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', '500', '300'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', '50A'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', '500', 'A=300', 'A=200'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', '500', 'A6=300'])
        params = Params(args)

def test_option_susceptibility():
    args = parse_args(['--popsize', 'A=500', 'B=300', '--susceptibility', 'A=0.8'])
    params = Params(args)
    assert sorted(list(params.groups.keys())) == ['A', 'B']

    assert params.susceptibility_multiplier_A == 0.8
    with pytest.raises(AttributeError):
        params.susceptibility_multiplier_B

    # group does not exist
    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=500', 'B=300', '--susceptibility', 'C=0.8'])
        params = Params(args)

    # not as multiplier
    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=500', 'B=300', '--susceptibility', '0.8'])
        params = Params(args)

def test_option_symptomatic_r0():
    args = parse_args(['--symptomatic-r0', '1.0'])
    params = Params(args)
    assert params.symptomatic_r0_low == 1.0
    assert params.symptomatic_r0_high == 1.0

    args = parse_args(['--symptomatic-r0', '1.2', '2.0'])
    params = Params(args)
    assert params.symptomatic_r0_low == 1.2
    assert params.symptomatic_r0_high == 2.0

    args = parse_args(['--popsize', 'A=500', 'B=300',
         '--symptomatic-r0', '1.5', 'A=0.8'])
    params = Params(args)
    assert sorted(list(params.groups.keys())) == ['A', 'B']
    assert params.symptomatic_r0_low == 1.5
    assert params.symptomatic_r0_high == 1.5
    assert params.symptomatic_r0_multiplier_A == 0.8

    # not float number
    with pytest.raises(ValueError):
        args = parse_args(['--symptomatic-r0', '1.0f'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--symptomatic-r0', '0.8a', '1.0'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--symptomatic-r0', '0.8', '1.0f'])
        params = Params(args)

    # more numbers
    with pytest.raises(ValueError):
        args = parse_args(['--symptomatic-r0', '0.8', '1.0', '2.0'])
        params = Params(args)

def test_option_asymptomatic_r0():
    args = parse_args(['--asymptomatic-r0', '1.0'])
    params = Params(args)
    assert params.asymptomatic_r0_low == 1.0
    assert params.asymptomatic_r0_high == 1.0

    args = parse_args(['--asymptomatic-r0', '1.2', '2.0'])
    params = Params(args)
    assert params.asymptomatic_r0_low == 1.2
    assert params.asymptomatic_r0_high == 2.0

    args = parse_args(['--popsize', 'A=500', 'B=300',
         '--asymptomatic-r0', '1.5', 'A=0.8'])
    params = Params(args)
    assert sorted(list(params.groups.keys())) == ['A', 'B']
    assert params.asymptomatic_r0_low == 1.5
    assert params.asymptomatic_r0_high == 1.5
    assert params.asymptomatic_r0_multiplier_A == 0.8

    # not float number
    with pytest.raises(ValueError):
        args = parse_args(['--asymptomatic-r0', '1.0f'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--asymptomatic-r0', '0.8a', '1.0'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--asymptomatic-r0', '0.8', '1.0f'])
        params = Params(args)

    # more numbers
    with pytest.raises(ValueError):
        args = parse_args(['--asymptomatic-r0', '0.8', '1.0', '2.0'])
        params = Params(args)

def test_option_inclubation_period():
    with pytest.raises(ValueError):
        args = parse_args(['--incubation-period', '1.0'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--incubation-period', '1.0', '2.0'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--incubation-period', 'uniform', '1.0', '2.0'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=100', 'B=200',
            '--incubation-period', 'C=0.8'])
        params = Params(args)

    args = parse_args(['--incubation-period', 'normal', '1.2', '2.0'])
    params = Params(args)
    assert params.incubation_period_loc == 1.2
    assert params.incubation_period_scale == 2.0

    args = parse_args(['--incubation-period', 'lognormal', '1.', '2.5'])
    params = Params(args)
    assert params.incubation_period_mean == 1.
    assert params.incubation_period_sigma == 2.5

    args = parse_args(['--popsize', 'A=500', 'B=300',
         '--incubation-period', 'A=0.8'])
    params = Params(args)
    assert sorted(list(params.groups.keys())) == ['A', 'B']
    assert params.incubation_period_multiplier_A == 0.8

    args = parse_args(['--popsize', 'A=500', 'B=300',
        '--incubation-period', 'normal', '5', '2', 'B=0.95', 'A=1.13'])
    params = Params(args)
    assert sorted(list(params.groups.keys())) == ['A', 'B']
    assert params.incubation_period_loc == 5.0
    assert params.incubation_period_scale == 2.0
    assert params.incubation_period_multiplier_A == 1.13
    assert params.incubation_period_multiplier_B == 0.95

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=500', 'B=300',
            '--incubation-period', 'normal', '5', '2.3p'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=500', 'B=300',
            '--incubation-period', 'normal', '5', '2.3p', 'A=1.4p'])
        params = Params(args)

    # negative multiplier
    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=500', 'B=300',
            '--incubation-period', 'A=-1'])
        params = Params(args)

def test_option_pre_quarantine():
    args = parse_args(['--pre-quarantine', '10'])
    params = Params(args)

    args = parse_args(['--popsize', 'A=10', 'B=10',
         '--pre-quarantine', '10', 'A1', 'B2'])
    params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=10', 'B=10',
             '--pre-quarantine', '10', '1'])
        params = Params(args)

    # index out of range
    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=10', 'B=10',
             '--pre-quarantine', '10', 'C1', 'C2'])
        params = Params(args)

    # missing quarantine time
    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=10', 'B=10',
             '--pre-quarantine', 'A1', 'A2'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', '100',
             '--pre-quarantine', '10', '101'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=10', 'B=10',
             '--pre-quarantine', '10', 'A10'])
        params = Params(args)

def test_option_infectors():
    args = parse_args(['--infectors', '10'])
    params = Params(args)

    args = parse_args(['--popsize', 'A=10', 'B=10',
         '--infectors', 'A1', 'B2'])
    params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=10', 'B=10',
             '--infectors', '1'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=10', 'B=10',
             '--infectors', 'C1', 'C2'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=10', 'B=10',
             '--infectors', 'A10'])
        params = Params(args)

def test_option_prop_asym_carriers():
    args = parse_args(['--prop-asym-carriers', '0.1'])
    params = Params(args)
    assert params.prop_asym_carriers_loc == 0.1
    assert params.prop_asym_carriers_scale == 0.0

    args = parse_args(['--prop-asym-carriers', '0.1', '0.3'])
    params = Params(args)
    assert params.prop_asym_carriers_loc == 0.2

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=10', 'B=10',
             '--prop-asym-carriers', 'C=1.2'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=10', 'B=10',
             '--prop-asym-carriers', 'A=-1.2'])
        params = Params(args)

    with pytest.raises(ValueError):
        args = parse_args(['--popsize', 'A=10', 'B=10',
             '--prop-asym-carriers', 'A=1.2f'])
        params = Params(args)

    args = parse_args(['--popsize', 'A=10', 'B=10',
            '--prop-asym-carriers', 'A=1.2'])
    params = Params(args)
    assert params.prop_asym_carriers_multiplier_A == 1.2

def test_main_default():
    main(['--jobs', '1', '--repeats', '100'])
    main(['--jobs', '1', '--repeats', '100', '--logfile', 'test.out'])
    assert os.path.isfile('test.out')
    main(['--analyze-existing-logfile', '--logfile', 'test.out'])

def test_main_symptomatic():
    main(['--jobs', '1', '--repeats', '100', '--symptomatic-r0', '1.0'])
    main(['--jobs', '1', '--repeats', '100', '--symptomatic-r0', '1.0', '2.0'])

def test_main_asymptomatic():
    main(['--jobs', '1', '--repeats', '100', '--asymptomatic-r0', '0.5'])
    main(['--jobs', '1', '--repeats', '100', '--asymptomatic-r0', '0.5', '1.5'])
    main(['--jobs', '1', '--repeats', '100', '--popsize', 'A=100', 'B=50',
        '--asymptomatic-r0', '0.5', '1.5', 'A=1.2', '--infector', 'A0'])

def test_main_pre_quarantine():
    main(['--jobs', '1', '--repeats', '100', '--pre-quarantine', '7'])
    main(['--jobs', '1', '--repeats', '100', '--pre-quarantine', '7', '1', '2'])
    main(['--jobs', '1', '--repeats', '100', '--popsize', 'A=100', 'B=200',
        '--pre-quarantine', '7', 'A2', 'A7',  '--infector', 'A2', 'A7'])

def test_main_incubation_period():
    main(['--jobs', '1', '--repeats', '100', '--incubation-period', 'lognormal', '1', '0.2'])
    main(['--jobs', '1', '--repeats', '100', '--incubation-period', 'normal', '1', '2'])




#!/usr/bin/env python
"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

requirements = ['tqdm', 'numpy', 'scipy']

setup_requirements = [
    'pytest-runner',
]

test_requirements = [
    'pytest>=3',
]

setup(
    author="Bo Peng",
    author_email='ben.bob@gmail.com',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Population-based Forward-time Simulator for the Outbreak of COVID-19",
    entry_points={
        'console_scripts': [
            'outbreak_simulator=covid19_outbreak_simulator.cli:main',
        ],
    },
    install_requires=requirements,
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords='covid19_outbreak_simulator',
    name='covid19-outbreak-simulator',
    packages=find_packages(
        include=['covid19_outbreak_simulator', 'covid19_outbreak_simulator.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/ictr/covid19-outbreak-simulator',
    version='0.2.2',
    zip_safe=False,
)



#!/usr/bin/env python
"""Tests for `covid19_outbreak_simulator` package."""

import pytest
import math
from scipy.stats import norm
import numpy as np


def test_params_set_mean(params):
    params.set('prop_asym_carriers', 'loc', 0.3)
    assert params.prop_asym_carriers_loc == 0.3


def test_prop_asym_carriers(params):
    params.set('prop_asym_carriers', prop='loc', value=0.25)
    params.set('prop_asym_carriers', prop='quantile_2.5', value=0.1)

    dist = norm(
        loc=params.prop_asym_carriers_loc,
        scale=params.prop_asym_carriers_scale)

    assert math.fabs(dist.cdf(0.1) - 0.025) < 0.001
    assert math.fabs(dist.cdf(0.4) - 0.975) < 0.001


def test_symptomatic_r0(params):
    params.set('symptomatic_r0', prop='low', value=1.4)
    params.set('symptomatic_r0', prop='high', value=2.8)

    assert params.symptomatic_r0_low == 1.4
    assert params.symptomatic_r0_high == 2.8


def test_asymptomatic_r0(params):
    params.set('asymptomatic_r0', prop='low', value=0.14)
    params.set('asymptomatic_r0', prop='high', value=0.28)

    assert params.asymptomatic_r0_low == 0.14
    assert params.asymptomatic_r0_high == 0.28


def test_draw_prop_asym_carriers(default_model):
    props = []
    for i in range(1000):
        props.append(default_model.draw_prop_asym_carriers())

    assert math.fabs(
        sum(props) / 1000 - default_model.params.prop_asym_carriers_loc) < 0.05


def test_drawn_is_asymptomatic(default_model):
    prop = default_model.draw_prop_asym_carriers()
    is_asymp = []
    for i in range(1000):
        is_asymp.append(default_model.draw_is_asymptomatic())

    assert math.fabs(sum(is_asymp) / 1000) - prop < 0.05


def test_drawn_random_r0(default_model):
    symp_r0 = []
    N = 10000
    for i in range(N):
        symp_r0.append(default_model.draw_random_r0(symptomatic=True))
    #
    assert math.fabs(sum(symp_r0) /
                     N) - (default_model.params.symptomatic_r0_low +
                           default_model.params.symptomatic_r0_high) / 2 < 0.005

    asymp_r0 = []
    for i in range(N):
        asymp_r0.append(default_model.draw_random_r0(symptomatic=False))
    #
    assert math.fabs(sum(asymp_r0) / N) - (
        default_model.params.asymptomatic_r0_low +
        default_model.params.asymptomatic_r0_high) / 2 < 0.005


def draw_random_incubation_period(default_model):
    symp_r0 = []
    N = 10000
    for i in range(N):
        symp_r0.append(default_model.draw_random_incubation_period())
    #
    assert math.fabs(sum(symp_r0) / N) - scipy.stats.lognorm.mean(
        s=1,
        loc=self.params.incubation_period_mean,
        scale=self.params.incubation_period_sigma) < 0.001


def test_get_symptomatic_transmission_probability(default_model):

    incu = default_model.draw_random_incubation_period()
    R0 = default_model.draw_random_r0(symptomatic=True)

    N = 10000
    r = []
    for i in range(N):
        x_grid, prob = default_model.get_symptomatic_transmission_probability(
            incu, R0)
        infected = np.random.binomial(1, prob, len(x_grid))
        r.append(sum(infected))
    #
    assert math.fabs(sum(r) / N) - R0 < 0.05


def test_get_asymptomatic_transmission_probability(default_model):

    R0 = default_model.draw_random_r0(symptomatic=False)

    N = 10000
    r = []
    for i in range(N):
        x_grid, prob = default_model.get_asymptomatic_transmission_probability(
            R0)
        infected = np.random.binomial(1, prob, len(x_grid))
        r.append(sum(infected))
    #
    assert math.fabs(sum(r) / N) - R0 < 0.05
